#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 25 15:42:37 2019

@author: ben
"""

from pystarworlds.Environment import Environment, Ambient, Physics
from pystarworlds.Event import Action, Perception
from pystarworlds.Agent import Body, Mind, Actuator, Sensor

class _Physics(Physics):
    pass

class _Ambient(Ambient):
    
    def __init__(self, agents):
        super(_Ambient, self).__init__(agents)
        self.ACTION_TEST = False
        
    def action_test(self):
        self.ACTION_TEST = True

class _Action(Action):
    executor = lambda ambient, action: ambient.action_test()

class _Perception(Perception):
    pass

class _Mind(Mind):
    
    def cycle(self):
        action = _Action()
        self.body.find_actuators(action)[0].attempt(action)
    
class _Actuator(Actuator):
    subscribe = [_Action]
    
    def __init__(self):
        super(_Actuator, self).__init__()
        self.ACTUATOR_TEST = False
    
    def attempt(self, action):
        self.ACTUATOR_TEST = True
        super(_Actuator, self).attempt(action)
        
class _Sensor(Sensor):
    subscribe = [_Perception]
    
    def __init__(self):
        super(_Sensor, self).__init__()
        self.SENSOR_TEST = False
        
    def notify(self, perception):
        self.SENSOR_TEST = True

actuator = _Actuator()
sensor = _Sensor()
agents = [Body(_Mind(), [actuator], [sensor])]
actions = [_Action]
physics = _Physics(actions)
ambient = _Ambient(agents)

e = Environment(physics,ambient)

e.simulate()


assert ambient.ACTION_TEST 
assert actuator.ACTUATOR_TEST
assert sensor.SENSOR_TEST


"""
@author: Nausheen Saba Shahid
@author: Benedict Wilkins
"""

#from ABC import abstractmethod

from .Identifiable import Identifiable
from .Event import Transient


class Mind:
       
    def __post_init__(self, body):
        self.body = body
        
    #@abstractmethod
    #def cycle(self):
    #    pass

#######################################

class Body(Identifiable):
    
    def __init__(self, mind, actuators, sensors):
        self.mind = mind
        #self.__name__ = name #ID IS ALREADY PART OF THE AGENT  - Identifiable
        if isinstance(actuators, dict) or isinstance(actuators, list) or isinstance(actuators, tuple):
            self._actuators = actuators
            if isinstance(actuators, dict):
                self.actuators = lambda : list(self._actuators.values())
            else:
                self.actuators = lambda : self._actuators
            for a in self.actuators():
                a.owner = self.ID
        else:
            raise ValueError("actuators must be a dict, list or tuple")
        if isinstance(sensors, dict) or isinstance(sensors, list) or isinstance(sensors, tuple):
            self._sensors = sensors
            if isinstance(sensors, dict):
                self.sensors = lambda : list(self._sensors.values())
            else:
                self.sensors = lambda : self._sensors

            for s in self.sensors():
                s.owner = self.ID
        else:
            raise ValueError("sensors must be a dict, list or tuple")
            
        self.mind.__post_init__(self)
        
    def find_actuators(self, action):
        return self.find(action, self.actuators())
        
    def find_sensors(self, perception):
        return self.find(perception, self.sensors())
    
    def find(self, event, collection):
        return [arg for arg in collection if type(event) in arg.subscribe]
                    
    def cycle(self):
        return self.mind.cycle()
    
    def __hash__(self):
        return self.ID.__hash__()
        
    def __eq__(self):
        return self.ID.__eq__()

class Actuator(Identifiable, Transient):
  
    def __init__(self):
        super(Actuator, self).__init__()
        self.owner = None
        
    def attempt(self, action):
        action.__post_init__(self.owner)
        super(Actuator, self).sink(action)

class Sensor(Identifiable, Transient):
    
    def __init__(self):
       super(Sensor, self).__init__()
       self.owner = None

    def notify(self, perception):
        super(Sensor, self).sink(perception)


"""
@author: Nausheen Saba Shahid
"""


# class for event
# class for source of event
from abc import  abstractmethod
from .Identifiable import Identifiable

class Source:
    
    def __iter__(self):
        return self
    
    def __next__(self):
        return self.source()
        
    @abstractmethod
    def source(self):
        pass
   
    # class for sink of event

class Sink:
    
    def __init__(self, buffer):
        self.buffer = buffer
    
    def sink(self, event):
        assert type(event) in type(self).subscribe
        self.buffer.append(event)
  
    # class for buffer of event
  
    
class Transient(Source, Sink):
    
    def __init__(self):
        Sink.__init__(self, [])
        Source.__init__(self)

    def __len__(self):
        return len(self.buffer)
    
    def __next__(self):
        if(len(self.buffer) == 0):
            raise StopIteration
        return self.source()
    
    def source(self):
       return self.buffer.pop()
    
    def sink(self, event):
        self.buffer.append(event)
        
    def setEmpty(self):
       self.buffer =[]
      
    def isEmpty(self):
        if(len(self.buffer)):
         return False
        else: 
         return True
     
class Event(Identifiable):
    
    def __init__(self):
        pass 
    
class Perception(Event):
    
    def __init__(self):
        super(Perception, self).__init__()
        
class Action(Event):
    
    def __init__(self):
        super(Action, self).__init__()
        self.__actor__ = None
        
    def __post_init__(self, actor):
        self.__actor__ = actor
        
class Executor:
    
    def __init__(self, _type):
        self._type = _type
        
    def __call__(self, env, action):
        pass
    
      


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 13:12:18 2019

@author: Benedict Wilkins
"""

import setuptools

setuptools.setup(name='pystarworlds',
      version='0.0.2',
      description='',
      url='https://github.com/dicelab-rhul/pystarworlds',
      author='Nausheen Saba, Benedict Wilkins, Joel Clarke',
      author_email='nausheen.shahid.2017@live.rhul.ac.uk, brjw@hotmail.co.uk, joel.clarke273@gmail.com',
      license='GNU General Public License v3 (GPLv3)',
      packages=setuptools.find_packages(),
      classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ])



"""
@author: Nausheen Saba Shahid
"""


from abc import ABC

class Identifiable(ABC):    #Identifiable is inherited from ABC
         
    def SimpleIDGenerator():  # Unique Id generator
        i = 0
        while(True):
            val = yield    
            yield val + str(i)
            i += 1
    
    IDGEN = SimpleIDGenerator()
    
    def __new__(typ, *args, **kwargs):   # new cannot be overrided by any sub class
        obj = super(Identifiable, typ).__new__(typ)
        next(Identifiable.IDGEN)
        obj.ID = Identifiable.IDGEN.send("@") #+ obj.__class__.__name__)
        return obj
    
    def __str__(self):       # equvilent to tostring
        return self.__class__.__name__ + self.ID
    
    def __unicode__(self):
        return self.__str__()
    
    def __repr__(self):
        return self.__str__()
    



"""
@author: Nausheen Saba Shahid
@author: Benedict Wilkins
"""

from .Identifiable import Identifiable
#from .Perception import Perception
from collections import defaultdict

class Environment(Identifiable):
    
    def __init__(self, physics, ambient, processes=[], *args):  
        self.physics = physics
        self.ambient = ambient
        #self.actions = actions
        #self.attempts = [] 
        self.processes = processes
        
        #move this to the physics ...?
        #self.physics.__post_init__(self.ambient.agents.values())        
            
#########################################################################
        
    def simulate(self, cycles=1):
        for i in range(cycles):
            self.evolveEnvironment()   

                  
    def evolveEnvironment(self):
        #allow all processes to do their thing
        for process in self.processes:
            process(self) #TODO fix this, processes should return events!

        agents = self.ambient.agents.values()
        for a in agents:
            a.cycle()
            
        attempts = [action for agent in agents for actuator in agent.actuators() for action in actuator]
        events = self.physics.execute(self, attempts)
    
    def __str__(self):
        return super().__str__() + "~"+  str(self.__ambient__)
    
    def __unicode__(self):
        return self.__str__()
    
    def __repr__(self):
        return self.__str__()
    
    
class Ambient(Identifiable):
    
    def __init__(self, agents, objects=[]):
        self.agents = {ag.ID:ag for ag in agents}
        self.objects = {obj.ID:obj for obj in objects} #what is this conceptually?  
    
class Physics(Identifiable):
         
    def __init__(self, actions): 
        self.subscriptions = defaultdict(list)
        self.actions = actions
        self.executors = {action:action.executor(action) for action in self.actions}
     
    def __post_init__(self, agents):
        sensors = [s for a in agents for s in a.sensors()]
        for sensor in sensors:
            for sub in sensor.subscribe:
                self.subscriptions[sub].append(sensor)

    def notify_agent(self, agent, event):
        #rethink this
        for sensor in agent.sensors():
            if type(event) in sensor.subscribe:
                sensor.notify(event)

    def execute(self, env, attempts):
        events = []
        for attempt in attempts:
            #print(attempt)
            #print(self.preconditions[type(attempt)])
           # if self.preconditions[type(attempt)](ambient, attempt): #this is checking a pre-condition
           new_events = self.executors[type(attempt)](env, attempt) #make changes to the ambient
           if new_events:
               events.extend(new_events)
        return events #events were generated by executing the actions
        
        
''' No.
   def execute(self,attempts,env,valid_actions,rule_factories,executeaction_factories):
       for act in attempts:
           if type(act) in valid_actions:
               for rule_factory in rule_factories:
                  if(type(act)==rule_factory._type) and rule_factory(env,act):
                   for executeaction_factory in executeaction_factories:
                      if (type(act)==executeaction_factory._type):     
                         executeaction_factory(env,act)
                        
'''          

class Process(Identifiable):
    
    def __init__(self):
        pass
    
    def __call__(self, ambient):
        pass #return a list of events
            
                
   
  
   











